/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <generic.h>
#include <outchannels.h>
#include <jutils.h>
#include <config.h>

/* I think a tighter bound could be:  (mt, March 2000)
 * MPEG1:
 *    num_samples*(bitrate/8)/samplerate + 4*1152*(bitrate/8)/samplerate + 512
 * MPEG2:
 *    num_samples*(bitrate/8)/samplerate + 4*576*(bitrate/8)/samplerate + 256
 */

/* mp3 encoder */

LameEnc::LameEnc() 
  : Profile("lame") {
  
  static struct settings lame_settings[] = {
    { "bitrate", (void*)&bps, cfgINT, "24" },
    { "frequency", (void*)&freq, cfgINT, "0" },
    { "mode", (void*)&mode, cfgINT, "3" },
    { "quality", (void*)&quality, cfgINT, "5" },
    { "lowpass", (void*)&lowpass, cfgINT, "0" },
    { "highpass", (void*)&highpass, cfgINT, "0" },
    { NULL, NULL, cfgNULL, NULL }
  };

  setup(lame_settings);

  initialized = false;
  encode_last = false;
}

/* note: num is number of samples in the L (or R) channel
   not the total number of samples in pcm[] */
bool LameEnc::encode(short int *pcm, int num) {    
  double now;
  encoded = 
    lame_encode_buffer_interleaved
    (enc_flags, pcm, num, buffer, ENCBUFFER_SIZE);
  
  if(encoded<0)
    switch(encoded) {
    case -1:
      error("lame encoder: buffer is too small");
      return false;
    case -2:
      error("lame encoder: malloc() problem");
      return false;
    case -3:
      error("lame encoder: lame_init_params() not called");
      return false;
    case -4:
      error("lame encoder: psycho acoustic problems (yes, shamanic indeed)");
      return false;
    default:
      error("lame encoder: unknown error :(");
      return false;
    }

  /* calcolates bitrate */
  bytes_accu += encoded;
  now = dtime();
  if((now-prev)>1) { /* if one second passed */
    bitrate = (bytes_accu<<3);
    bytes_accu = 0;
    prev = now;
  }
  return true;
}

void LameEnc::flush() {
  encoded = lame_encode_flush(enc_flags,buffer,ENCBUFFER_SIZE);
}

void LameEnc::flush_nogap() {
  encoded = lame_encode_flush_nogap(enc_flags,buffer,ENCBUFFER_SIZE);
}

void LameEnc::init() {
  enc_flags = lame_init();

  lame_set_num_samples(enc_flags,MIX_CHUNK<<2); /* we get 1024 samples */
  lame_set_num_channels(enc_flags,2); /* the mixed input stream is stereo */
  /*  lame_set_in_samplerate(enc_flags,SAMPLE_RATE); the default mixed input stream is 44khz */
  lame_set_error_protection(enc_flags,1); /* 2 bytes per frame for a CRC checksum */

  default_profile();
}

bool LameEnc::apply_profile() {

  lame_set_compression_ratio(enc_flags,0);
  lame_set_brate(enc_flags,bps);


  lame_set_out_samplerate(enc_flags,freq);

  lame_set_mode(enc_flags,(enum MPEG_mode_e)mode); /* 0,1,2,3 stereo,jstereo,dual channel,mono */
  lame_set_quality(enc_flags,quality); /* 1..9 1=best 9=worst (suggested: 2, 5, 7) */
  
  /* lame chooses for us frequency filtering when values are 0 */
  lame_set_lowpassfreq(enc_flags,lowpass);
  lame_set_highpassfreq(enc_flags,highpass);


  
  
  if(lame_init_params(enc_flags)<0) {
    error("lame initialization failed");
    initialized = false;
  } else initialized = true;

  freq = lame_get_out_samplerate(enc_flags);
  
  /* save the active section here! */
  write_profile("last_active");

  return(initialized);
}

char *LameEnc::version() {
  sprintf(ver,"Lame %s",get_lame_version());
  return(ver);
}

void LameEnc::close() {
  act("closing lame encoder");
  lame_close(enc_flags);
}

LameEnc::~LameEnc() {
  if(initialized) close();
}

/* IcecastOut */

IcecastOut::IcecastOut()
  : Profile("ice") {

  static struct settings ice_settings[] = {
    { "ip_address", (void*)&ip, cfgSTR, "127.0.0.1" },
    { "port", (void*)&port, cfgINT, "8000" },
    { "pass", (void*)&pass, cfgSTR, "hackme" },
    { "mount", (void*)&mount, cfgSTR, "live" },
    { "bps", (void*)&bps, cfgINT, "24" },
    { "login", (void*)&login, cfgINT, "0" },
    { "name", (void*)&name, cfgSTR, "Streaming with MuSE" },
    { "url", (void*)&url, cfgSTR, "http://muse.dyne.org" },
    { "desc", (void*)&desc, cfgSTR, "Free Software Multiple Streaming Engine" },
    { NULL, NULL, cfgNULL, NULL }
  };
  setup(ice_settings);

  initialized = false;
  running = false;
}

IcecastOut::~IcecastOut() {
  func("IcecastOut::~IcecastOut");
  if(running) {
    act("closing icecast");
    stop();
  }
}

bool IcecastOut::init() {
  default_profile();
  return initialized = true;
}

bool IcecastOut::start() {
  int res;

  if(running) return(true);
  
  if(initialized) {

    notice("Contacting icecast server %s on port %u",ip,port);
    
    res = shout_connect(&ice_conn);
    if(res) { 
      act("allright, we are streaming!");
      running = true;
    } else {
      shout_strerror(&ice_conn,ice_conn.error,errormsg,ERRORMSG);
      error("%s",errormsg);
      act("streaming aborted, check settings for your server");
      running = false;
    }

    return(running);
  } else
    return(false); /* icecast connection not initialized */
}

void IcecastOut::stop() {
  if(running) {
    shout_disconnect(&ice_conn);
    running = false;
  }
}

bool IcecastOut::apply_profile() {
  stop();

  shout_init_connection(&ice_conn);
  ice_conn.ip = ip;
  ice_conn.port = port;
  ice_conn.password = pass;
  ice_conn.mount = mount;
  ice_conn.bitrate = bps*1000;
  ice_conn.name = name;
  ice_conn.url = url;
  ice_conn.description = desc;
  ice_conn.icy_compat = login;
  ice_conn.ispublic = 1;

  /* save the active section here! */
  write_profile("last_active");

  return start();
}

bool IcecastOut::set_server(char *hostname) {

  struct sockaddr_in server_addr;
  struct hostent *h;
  
  memset(&server_addr, 0, sizeof(struct sockaddr_in));
  server_addr.sin_family = AF_INET;
  
#ifdef HAVE_INET_ATON
  inet_aton(hostname,&server_addr.sin_addr);
#else
  server_addr.sin_addr.s_addr=inet_addr(hostname);
#endif
  
  if(!(h = gethostbyname(hostname)))
    return(false);
  
  server_addr.sin_family = AF_INET;
  memcpy(&server_addr.sin_addr, h->h_addr, h->h_length);    
  
  sprintf(ip,"%s",
	  inet_ntoa(*(struct in_addr *) &server_addr.sin_addr.s_addr));
  func("%s resolved to %s",hostname, ip);
  
  return(true);
}

int IcecastOut::send(unsigned char *buf, unsigned int enc) {
  if(!running) return(0);
  if(enc<=0) return(0);

  int res;
  sync();
  res = shout_send_data(&ice_conn, buf, enc);
  if(!res) {
    shout_strerror(&ice_conn,ice_conn.error,errormsg,ERRORMSG);
    warning("%s",errormsg);
    return(-1);
  }

  return(res);
}

void IcecastOut::sync() {
  shout_sleep(&ice_conn);
}

/* Mp3fileOut */

Mp3fileOut::Mp3fileOut() { 
  fd = NULL;
}

bool Mp3fileOut::openmp3(char *file) {
  int num = 0;
  struct stat st;
  char temp[MAX_PATH_SIZE];

  closemp3();  

  snprintf(temp,MAX_PATH_SIZE,"%s",file);

  while(stat(temp, &st) != -1) {
    /* file EXIST */
    num++;
    snprintf(temp,MAX_PATH_SIZE,"%s.%i",file,num);
  }
  
  fd = fopen(temp,"w"); /* creates a file writeonly nonblocking (-rw-rw-r--) */
  if(!fd) {
    error("Mp3fileOut::openmp3(%s) : can't open file",temp);
    act("%s",strerror(errno));
    return(false);
  }

  strncpy(filename,temp,MAX_PATH_SIZE);
  notice("dumping to mp3 file %s",filename);
  return(true);
}

int Mp3fileOut::writemp3(unsigned char *buf, unsigned int enc) {
  int res = -1;
  
  if(fd&&(enc>0)) {
    syncmp3();
    res = fwrite(buf,1,enc,fd);
  }
  return(res);
}

void Mp3fileOut::closemp3() {
  if(fd) {
    syncmp3();
    act("closing mp3 record file");
    fclose(fd);
    fd = NULL;
  }
}
 
bool Mp3fileOut::syncmp3() {
  if(fflush(fd)!=0) return(false);
  else return(true);
}

Mp3fileOut::~Mp3fileOut() {
  func("Mp3fileOut::~Mp3fileOut()");
  closemp3();
}
